/*
 * Copyright 2020-2022 Neural Layer
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     https://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.roboquant.common

import java.text.DecimalFormat

interface Summarizable {

    fun summary(singleCurrency: Boolean = true): Summary
}

/**
 * Summary allows to represent nested data into a tree like format. This is supported by several of the components
 * in roboquant if you invoke their summary() method.
 *
 * @property content The content for the summary
 * @constructor Create new Summary
 */
class Summary(val content: String) {

    private val children = mutableListOf<Summary>()
    private val decimalFormatter = DecimalFormat(decimalPattern)

    companion object {
        var decimalPattern = "#.000"
        var sep = ": "
    }

    /**
     * Add a new summary as a child
     *
     * @param child
     */
    fun add(child: Summary) = children.add(child)

    /**
     * Add a [label] with a numerical [value]. When adding a Float or Double, the decimal formatter will be used to
     * format it. Other numbers will be presented using the toString() method.
     */
    fun add(label: String, value: Number) {
        when (value) {
            is Double -> children.add(Summary("$label$sep${decimalFormatter.format(value)}"))
            is Float -> children.add(Summary("$label$sep${decimalFormatter.format(value)}"))
            else -> children.add(Summary("$label$sep$value"))
        }
    }

    /**
     * Add a [label] and [value] to this summary
     */
    fun add(label: String, value: Any?) = children.add(Summary("$label$sep$value"))


    /**
     * Add a [label] to this summary
     */
    fun add(label: String) = children.add(Summary(label))


    /**
     * To string
     */
    override fun toString(): String {
        val buffer = StringBuilder()
        generate(buffer, "", "")
        return buffer.toString()
    }

    private fun generate(
        buffer: StringBuilder,
        prefix: String,
        childrenPrefix: String,
    ) {
        buffer.append(prefix)
        buffer.append(content)
        buffer.append('\n')
        val it = children.iterator()
        while (it.hasNext()) {
            val next = it.next()
            if (it.hasNext()) {
                next.generate(buffer, "$childrenPrefix├── ", "$childrenPrefix│   ")
            } else {
                next.generate(buffer, "$childrenPrefix└── ", "$childrenPrefix    ")
            }
        }
    }
}